using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using XNA4;
using Particle3D;

namespace XNA4Particle
{
    /// <summary>
    /// This is the main type for your game
    /// </summary>
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        SpriteFont font;

        // WVP
        float aspectRatio;// The aspect ratio determines how to scale 3d to 2d projection.
        Matrix world;
        Matrix view;
        Matrix projection;

        /// <summary>
        /// input component
        /// </summary>
        static InputHandler inputs;
        /// <summary>
        /// FPS component
        /// </summary>
        static FpsCounter fpsCounter;

        /// <summary>
        /// model node component
        /// </summary>
        ModelNode[] modelList = new ModelNode[3];
        // Set the position of the model in world space, and set the rotation.
        Vector3 modelPosition = Vector3.Zero;
        float modelRotation = 0.0f;
        float modelRotationPitch = 0.0f;
        // Set the velocity of the model, applied each frame to the model's position.
        Vector3 modelVelocity = Vector3.Zero;

        /// <summary>
        /// particle system component
        /// </summary>
        // This sample uses five different particle systems.
        ParticleSystem explosionParticles;
        ParticleSystem explosionSmokeParticles;
        ParticleSystem projectileTrailParticles;
        ParticleSystem smokePlumeParticles;
        ParticleSystem fireParticles;
        ParticleSystem rainParticles;

        // particle state
        enum ParticleState
        {
            Explosions,
            SmokePlume,
            RingOfFire,
        };
        ParticleState currentState = ParticleState.Explosions;

        // The explosions effect works by firing projectiles up into the
        // air, so we need to keep track of all the active projectiles.
        List<Projectile> projectiles = new List<Projectile>();
        TimeSpan timeToNextProjectile = TimeSpan.Zero;
        // Random number generator for the fire effect.
        Random random = new Random();
        
        // camera state.
        float cameraArc = -5;
        float cameraRotation = 0;
        float cameraDistance = 500;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            // TODO: Add your initialization logic here
            // WVP
            world = Matrix.CreateRotationY(modelRotation) * Matrix.CreateRotationX(modelRotationPitch) * Matrix.CreateTranslation(modelPosition);
            view = Matrix.CreateTranslation(0, -25, 0) *
                   Matrix.CreateRotationY(MathHelper.ToRadians(cameraRotation)) *
                   Matrix.CreateRotationX(MathHelper.ToRadians(cameraArc)) *
                   Matrix.CreateLookAt(new Vector3(0, 0, -cameraDistance), Vector3.Zero, Vector3.Up);
            aspectRatio = (float)graphics.GraphicsDevice.Viewport.Width / (float)graphics.GraphicsDevice.Viewport.Height;
            projection = Matrix.CreatePerspectiveFieldOfView(MathHelper.PiOver4, aspectRatio, 1.0f, 1000.0f);

            /// <summary>
            /// inputs component
            /// </summary>
            inputs = new InputHandler(this);
            Components.Add(inputs);
            /// <summary>
            /// fps component
            /// </summary>
            fpsCounter = new FpsCounter(this);
            Components.Add(fpsCounter);

            modelList[0] = new ModelNode(this, "Model\\grid", new Vector3(0, -50, 0), new Vector3(1.0f, 1.0f, 1.0f), Quaternion.Identity);
            modelList[0].World = world;
            modelList[0].View = view;
            modelList[0].Projection = projection;
            Components.Add(modelList[0]);
            modelList[1] = new ModelNode(this, "Model\\Ship", new Vector3(-100, 0, 0), new Vector3(0.1f, 0.1f, 0.1f), Quaternion.Identity); //Quaternion.CreateFromAxisAngle(new Vector3(0, 1, 0), MathHelper.Pi));
            modelList[1].World = world;
            modelList[1].View = view;
            modelList[1].Projection = projection;
            Components.Add(modelList[1]);
            modelList[2] = new ModelNode(this, "Model\\p1_wedge", new Vector3(100, 0, 0), new Vector3(0.1f, 0.1f, 0.1f), Quaternion.Identity); //Quaternion.CreateFromAxisAngle(new Vector3(0, 1, 0), MathHelper.Pi));
            modelList[2].World = world;
            modelList[2].View = view;
            modelList[2].Projection = projection;
            Components.Add(modelList[2]);
            
            // Construct our particle system components.
            explosionParticles = new ExplosionParticleSystem(this, Content);
            explosionSmokeParticles = new ExplosionSmokeParticleSystem(this, Content);
            projectileTrailParticles = new ProjectileTrailParticleSystem(this, Content);
            smokePlumeParticles = new SmokePlumeParticleSystem(this, Content);
            fireParticles = new FireParticleSystem(this, Content);
            // Set the draw order so the explosions and fire
            // will appear over the top of the smoke.
            smokePlumeParticles.DrawOrder = 100;
            explosionSmokeParticles.DrawOrder = 200;
            projectileTrailParticles.DrawOrder = 300;
            explosionParticles.DrawOrder = 400;
            fireParticles.DrawOrder = 500;
            // Register the particle system components.
            Components.Add(explosionParticles);
            Components.Add(explosionSmokeParticles);
            Components.Add(projectileTrailParticles);
            Components.Add(smokePlumeParticles);
            Components.Add(fireParticles);
            Components.Add(rainParticles);
            
            base.Initialize();
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
            // Create a new SpriteBatch, which can be used to draw textures.
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // TODO: use this.Content to load your game content here
            font = Content.Load<SpriteFont>("Font\\font");
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
            // TODO: Unload any non ContentManager content here
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            // Allows the game to exit
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // get some input.
            UpdateInput(gameTime);

            // update particle
            UpdateParticle(gameTime);

            base.Update(gameTime);
        }

        protected void UpdateInput(GameTime gameTime)
        {
            float speed = 5.0f;
            if (inputs.PauseGame)
            {
                this.Exit();
            }

            // model poosition & orientation reset
            if (inputs.IsKeyPressed(Keys.Space))
            {
                modelPosition = Vector3.Zero;
                modelVelocity = Vector3.Zero;
                modelRotation = 0.0f;
                modelRotationPitch = 0.0f;
            }

            // model position
            if (inputs.IsKeyDown(Keys.I))
            {
                modelPosition += Vector3.Up * speed;
                System.Diagnostics.Trace.WriteLine("modelPosition=" + modelPosition);
            }
            else if (inputs.IsKeyDown(Keys.K))
            {
                modelPosition += Vector3.Down * speed;
                System.Diagnostics.Trace.WriteLine("modelPosition=" + modelPosition);
            }
            else if (inputs.IsKeyDown(Keys.L))
            {
                modelPosition += Vector3.Right * speed;
                System.Diagnostics.Trace.WriteLine("modelPosition=" + modelPosition);
            }
            else if (inputs.IsKeyDown(Keys.J))
            {
                modelPosition += Vector3.Left * speed;
                System.Diagnostics.Trace.WriteLine("modelPosition=" + modelPosition);
            }
            else if (inputs.IsKeyDown(Keys.O))
            {
                modelPosition += Vector3.Forward * speed;
            }
            else if (inputs.IsKeyDown(Keys.P))
            {
                modelPosition += Vector3.Backward * speed;
            }
            // model orientation
            if (inputs.IsKeyDown(Keys.A))
            {
                modelRotation += 0.10f;
                //System.Diagnostics.Trace.WriteLine("modelRotation=" + modelRotation);
            }
            else if (inputs.IsKeyDown(Keys.D))
            {
                modelRotation -= 0.10f;
                //System.Diagnostics.Trace.WriteLine("modelRotation=" + modelRotation);
            }
            else if (inputs.IsKeyDown(Keys.W))
            {
                modelRotationPitch += 0.10f;
                //System.Diagnostics.Trace.WriteLine("modelRotation=" + modelRotation);
            }
            else if (inputs.IsKeyDown(Keys.S))
            {
                modelRotationPitch -= 0.10f;
                //System.Diagnostics.Trace.WriteLine("modelRotation=" + modelRotation);
            }

            // Check for changing the active particle effect.
            if (inputs.IsKeyPressed(Keys.X))
            {
                currentState++;

                if (currentState > ParticleState.RingOfFire)
                    currentState = 0;
            }

            //modelRotation -= inputs.MouseMoved.X * 0.01f;
            //modelRotationPitch -= inputs.MouseMoved.Y * 0.01f;

            UpdateCamera(gameTime);
        }

        /// <summary>
        /// Handles input for moving the camera.
        /// </summary>
        void UpdateCamera(GameTime gameTime)
        {
            float time = (float)gameTime.ElapsedGameTime.TotalMilliseconds;

            // camera reset
            if (inputs.IsKeyPressed(Keys.Space))
            {
                cameraArc = -5;
                cameraRotation = 0;
                cameraDistance = 500;
            }

            // input to rotate the camera up and down around the model.
            if (inputs.IsKeyDown(Keys.Up))
            {
                cameraArc += time * 0.025f;
            }
            else if (inputs.IsKeyDown(Keys.Down))
            {
                cameraArc -= time * 0.025f;
            }
            // limit the arc movement.
            if (cameraArc > 90.0f)
                cameraArc = 90.0f;
            else if (cameraArc < -90.0f)
                cameraArc = -90.0f;

            // input to rotate the camera around the model.
            if (inputs.IsKeyDown(Keys.Right))
            {
                cameraRotation += time * 0.05f;
            }
            else if (inputs.IsKeyDown(Keys.Left))
            {
                cameraRotation -= time * 0.05f;
            }

            // input to zoom camera in and out.
            if (inputs.IsKeyDown(Keys.PageUp))
                cameraDistance += time * 0.25f;
            else if (inputs.IsKeyDown(Keys.PageDown))
                cameraDistance -= time * 0.25f;
            // limit the camera distance.
            if (cameraDistance > 900)
                cameraDistance = 900;
            else if (cameraDistance < 100)
                cameraDistance = 100;
        }

        void UpdateParticle(GameTime gameTime)
        {            
            switch (currentState)
            {
                case ParticleState.Explosions:
                    UpdateExplosions(gameTime);
                    break;
                case ParticleState.SmokePlume:
                    UpdateSmokePlume();
                    break;
                case ParticleState.RingOfFire:
                    UpdateFire();
                    break;
            }
            UpdateProjectiles(gameTime);
        }

        /// <summary>
        /// Helper for updating the explosions effect.
        /// </summary>
        void UpdateExplosions(GameTime gameTime)
        {
            timeToNextProjectile -= gameTime.ElapsedGameTime;

            if (timeToNextProjectile <= TimeSpan.Zero)
            {
                // Create a new projectile once per second. The real work of moving
                // and creating particles is handled inside the Projectile class.
                projectiles.Add(new Projectile(explosionParticles,
                                               explosionSmokeParticles,
                                               projectileTrailParticles));

                timeToNextProjectile += TimeSpan.FromSeconds(1);
            }
        }


        /// <summary>
        /// Helper for updating the list of active projectiles.
        /// </summary>
        void UpdateProjectiles(GameTime gameTime)
        {
            int i = 0;

            while (i < projectiles.Count)
            {
                if (!projectiles[i].Update(gameTime))
                {
                    // Remove projectiles at the end of their life.
                    projectiles.RemoveAt(i);
                }
                else
                {
                    // Advance to the next projectile.
                    i++;
                }
            }
        }


        /// <summary>
        /// Helper for updating the smoke plume effect.
        /// </summary>
        void UpdateSmokePlume()
        {
            // This is trivial: we just create one new smoke particle per frame.
            smokePlumeParticles.AddParticle(Vector3.Zero, Vector3.Zero);
        }

        /// <summary>
        /// Helper for updating the fire effect.
        /// </summary>
        void UpdateFire()
        {
            const int fireParticlesPerFrame = 20;

            // Create a number of fire particles, randomly positioned around a circle.
            for (int i = 0; i < fireParticlesPerFrame; i++)
            {
                fireParticles.AddParticle(RandomPointOnCircle(), Vector3.Zero);
            }

            // Create one smoke particle per frmae, too.
            smokePlumeParticles.AddParticle(RandomPointOnCircle(), Vector3.Zero);
        }

        /// <summary>
        /// Helper used by the UpdateFire method. Chooses a random location
        /// around a circle, at which a fire particle will be created.
        /// </summary>
        Vector3 RandomPointOnCircle()
        {
            const float radius = 30;
            const float height = 40;

            double angle = random.NextDouble() * Math.PI * 2;

            float x = (float)Math.Cos(angle);
            float y = (float)Math.Sin(angle);

            return new Vector3(x * radius, y * radius + height, 0);
        }

        Vector3 GetRandomVector3(Vector3 min, Vector3 max)
        {
            return (new Vector3(
                GetRandomFloat(min.X, max.X),
                GetRandomFloat(min.Y, max.Y),
                GetRandomFloat(min.Z, max.Z)));
        }

        float GetRandomFloat(float min, float max)
        {
            return (((float)random.NextDouble() * (max - min)) + min);
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.CornflowerBlue);

            // TODO: Add your drawing code here
            view = Matrix.CreateTranslation(0, -25, 0) *
              Matrix.CreateRotationY(MathHelper.ToRadians(cameraRotation)) *
              Matrix.CreateRotationX(MathHelper.ToRadians(cameraArc)) *
              Matrix.CreateLookAt(new Vector3(0, 0, -cameraDistance),
                                  new Vector3(0, 0, 0), Vector3.Up);

            // model
            modelList[0].View = view;
            modelList[1].View = view;
            modelList[1].World = Matrix.CreateRotationY(modelRotation) * Matrix.CreateRotationX(modelRotationPitch) * Matrix.CreateTranslation(modelPosition);
            modelList[2].View = view;
            modelList[2].World = Matrix.CreateRotationY(modelRotation) * Matrix.CreateRotationX(modelRotationPitch) * Matrix.CreateTranslation(modelPosition);

            // particle
            DrawParticle(view, projection);

            // text
            DrawMessage();

            base.Draw(gameTime);
        }

        void DrawParticle(Matrix view, Matrix projection)
        {
            // Pass camera matrices through to the particle system components.
            explosionParticles.SetCamera(view, projection);
            explosionSmokeParticles.SetCamera(view, projection);
            projectileTrailParticles.SetCamera(view, projection);
            smokePlumeParticles.SetCamera(view, projection);
            fireParticles.SetCamera(view, projection);
        }

        /// <summary>
        /// Helper for drawing our message text.
        /// </summary>
        void DrawMessage()
        {
            string message = string.Format("Current effect: {0}!!!\n Hit the X button to switch.", currentState);

            spriteBatch.Begin();
            spriteBatch.DrawString(font, message, new Vector2(50, 50), Color.White);
            spriteBatch.End();
        }

    }
}
