﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace XNA4
{
    public class ModelNode : DrawableGameComponent
    {
        /// <summary>
        /// model object/name
        /// </summary>
        Model mesh;
        string name;

        /// <summary>
        /// model position/scale/rotation
        /// </summary>
        public Vector3 Position;
        public Vector3 Scale;
        public Quaternion Rotation;

        /// <summary>
        /// model matrix
        /// </summary>
        public Matrix World;
        public Matrix View;
        public Matrix Projection;

        // bounding volume
        BoundingBox volume;
        //public BoundingBox Bounds;

        /// <summary>
        /// ctor
        /// </summary>
        public ModelNode(Game game, string modelAssetName)
            : this(game, modelAssetName, Vector3.Zero, Vector3.One, Quaternion.Identity)
        {
        }

        public ModelNode(Game game, string modelAssetName, Vector3 position, Vector3 scale, Quaternion rotate)
            : base(game)
        {
            // model load
            this.name = modelAssetName;
            if (game.Content != null && this.mesh == null && this.name != string.Empty)
                this.mesh = game.Content.Load<Model>(this.name);

            // model transformation
            this.Position = position;
            this.Scale = scale;
            this.Rotation = rotate;
            
            // bounding volume     
            this.volume = UpdateBoundingBox();
        }

        /// <summary>
        /// Allows the game component to perform any initialization it needs to before starting
        /// to run.  This is where it can query for any required services and load content.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();
        }

        /// <summary>
        /// Allows the game component to update itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        public override void Update(GameTime gameTime)
        {
        }

        // model's bounding box
        protected BoundingBox UpdateBoundingBox() 
        {     
            // initialize minimum and maximum corners of the bounding box to max and min values     
            Vector3 min = new Vector3(float.MaxValue, float.MaxValue, float.MaxValue);     
            Vector3 max = new Vector3(float.MinValue, float.MinValue, float.MinValue);      
            // For each mesh of the model     
            foreach (ModelMesh mesh in this.mesh.Meshes)     
            {         
                foreach (ModelMeshPart meshPart in mesh.MeshParts)         
                {             
                    // Vertex buffer parameters             
                    int vertexStride = meshPart.VertexBuffer.VertexDeclaration.VertexStride;             
                    int vertexBufferSize = meshPart.NumVertices * vertexStride;              
                    // Get vertex data as float             
                    float[] vertexData = new float[vertexBufferSize / sizeof(float)];             
                    meshPart.VertexBuffer.GetData<float>(vertexData);              
                    // Iterate through vertices (possibly) growing bounding box, all calculations are done in world space             
                    for (int i = 0; i < vertexBufferSize / sizeof(float); i += vertexStride / sizeof(float))             
                    {                 
                        Vector3 transformedPosition = Vector3.Transform(new Vector3(vertexData[i], vertexData[i + 1], vertexData[i + 2]), World);                  
                        min = Vector3.Min(min, transformedPosition);                 
                        max = Vector3.Max(max, transformedPosition);             
                    }         
                }     
            }      
            // Create and return bounding box     
            return new BoundingBox(min, max); 
        }

        public override void Draw(GameTime gameTime)
        {
            if (this.mesh == null) return;

            // save current states
            BlendState bs = GraphicsDevice.BlendState;
            DepthStencilState dss = GraphicsDevice.DepthStencilState;
            // set for transparencies
            //GraphicsDevice.BlendState = BlendState.AlphaBlend;
            GraphicsDevice.BlendState = BlendState.Opaque;
            GraphicsDevice.DepthStencilState = DepthStencilState.Default;
            GraphicsDevice.SamplerStates[0] = SamplerState.LinearWrap;

            foreach (ModelMesh mesh in this.mesh.Meshes)
            {
                Matrix world = Matrix.CreateScale(this.Scale) * Matrix.CreateFromQuaternion(this.Rotation) * this.World * Matrix.CreateTranslation(this.Position);

                // this is where the mesh orientation is set, as well as our camera and projection
                foreach (BasicEffect effect in mesh.Effects)
                {
                    effect.EnableDefaultLighting();
                    effect.Alpha = 1.0f;
                    effect.World = world;
                    effect.View = this.View;
                    effect.Projection = this.Projection;
                }
                // draw the mesh, will use the effects set above.
                mesh.Draw();
            }
            // reset states
            GraphicsDevice.BlendState = bs;
            GraphicsDevice.DepthStencilState = dss;

            // draw bounding box
            DrawBoundingBox();
        }

        // draw bounding box
        public virtual void DrawBoundingBox()
        {
        }

    }
}